/*
   salias.c
  
   Copyright (C) 2003 Frank J. Tobin, ftobin@neverending.org

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of the
   License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA or see http://www.gnu.org/copyleft/lesser.html

 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <glib.h>

#define COPYRIGHT "Copyright (C) 2003 Frank Tobin <ftobin@neverending.org>"
#define REVISION  "$Id: salias.c,v 1.7 2003/03/31 06:30:46 ftobin Exp $"

#define subhome_fn "/.salias"

const char  *desired_context = NULL;
char       **global_argv;
int          global_argc;
int          cfg_lineno;

gboolean debug              = FALSE;
gboolean in_context         = FALSE;
gboolean in_desired_context = FALSE;

GString *cfg_fullpath = NULL;

/* alias is a list of alias lists, which are lists of strings,
   the first element of which is the alias name, followed by the
   alias values */
GSList *aliases        = NULL;
GSList *rep_options    = NULL;

#define slist_push(list, data)    g_slist_prepend(list, data)
GSList *slist_pop(GSList *list, void **popped);

int main (int argc, char **argv);
void usage(void);
void version(void);

void exec_it(void);
void xexecvp(const char *file, char *const argv[]);

GSList *argv_to_gslist(char **argv);
GArray *gslist_to_array(GSList *list);

GSList *get_subs(const char *alias);
GSList *get_default_rep_options(const char *name);
GSList *replace_args(GSList *argslist);
gboolean is_rep_option(const char *arg);

void print_dump(void);
void gfunc_printer(gpointer data, gpointer user_data);
void gfunc_innerprinter(gpointer data, gpointer user_data);
void printlist(GSList *list);

void handle_line(GString *s);
void handle_config_argv(int argc, char **argv);
void cfg_parse_error(const char *s);
void cfg_warning(const char *s);

void dummy_log(const gchar    *log_domain,
	       GLogLevelFlags  log_level,
	       const           gchar *message,
	       gpointer        user_data);


void usage(void)
{
    g_printerr("usage: salias [-d] [-h] [-v] [-f config] [-c context] program [args ...]\n");
    exit(EXIT_FAILURE);
}


void version(void)
{
    g_print("%s\n%s\n%s\nPlease send bug reports to %s\n",
	    PACKAGE_STRING,
	    REVISION,
	    COPYRIGHT,
	    PACKAGE_BUGREPORT
	);
    exit(EXIT_SUCCESS);
}


int main (int argc, char *argv[])
{
    int c;
    GIOChannel *f;
    GError     *err;
    GString    *readbuf;
    GIOStatus   status;
    
    g_log_set_handler(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, dummy_log, NULL);
    g_log_set_always_fatal(G_LOG_LEVEL_CRITICAL);

    while (1)
    {
	c = getopt(argc, argv, "+dvhc:f:");
	if (c == -1)
	{
	    break;
	}
	
	switch (c)
	{
	case 'd':
	    debug = TRUE;
	    break;
	case 'h':
	    usage();
	    g_assert_not_reached();
	case 'v':
	    version();
	    g_assert_not_reached();
	case 'c':
	    desired_context = optarg;
	    break;
	case 'f':
	    cfg_fullpath = g_string_new(optarg);
	    break;
	case '?':
	    exit(EXIT_FAILURE);
	}
    }

    if (argc - optind < 1)
    {
	usage();
    }
    
    if (desired_context == NULL)
    {
	desired_context = g_path_get_basename(argv[optind]);
    }
    
    global_argv = argv;
    global_argc = argc;

    if (cfg_fullpath == NULL)
    {
	cfg_fullpath = g_string_new(g_getenv("HOME"));
	cfg_fullpath = g_string_append(cfg_fullpath, subhome_fn);
	
	/* no config? just go ahead and exec */
	if (!g_file_test(cfg_fullpath->str, G_FILE_TEST_EXISTS))
	{
	    exec_it();
	}
    }
    
    err = NULL;
    if ((f = g_io_channel_new_file(cfg_fullpath->str, "r", &err)) == NULL)
    {
	g_assert(err != NULL);
	g_warning("error opening %s: %s",
		  cfg_fullpath->str, err->message);
	exec_it();
    }
    

    readbuf = g_string_new(NULL);
    for (cfg_lineno = 1; ; cfg_lineno++)
    {
	err = NULL;
	if ((status = g_io_channel_read_line_string(f, readbuf,
						    NULL, &err))
	    == G_IO_STATUS_EOF)
	{
	    break;
	}
	
	if (err != NULL)
	{
	    g_critical("error reading %s: %s",
		       cfg_fullpath->str, err->message);
	    g_assert_not_reached();
	}
	
	/* g_print("read in: %s", readbuf->str); */
	handle_line(readbuf);
    }
    
    /* if we got here, then we didn't find a matching context */
    exec_it();
    g_assert_not_reached();
    return EXIT_FAILURE;
}



/* note: this routine will destroy argslist */
GSList *replace_args(GSList *argslist)
{
    char   *arg;
    char   *rep_option  = NULL;
    GSList *newargslist = NULL;
    GSList *subs;
    GSList *el;

    /* we're going to treat argslist like a stack, with the top of
       the stack being the head of the list */
    /* we're also going to treat newargslist similarly like
       a 'backwards' (head at left) stack */
    while (argslist != NULL)
    {
	argslist = slist_pop(argslist, (void **)&arg);
	/* g_print("popped arg: %s\n", arg); */
	g_assert(arg != NULL);
	
	/* we reverse so that when we push back onto 
	   argslist one by one, they go on in the right order */
	subs = g_slist_reverse(g_slist_copy(get_subs(arg)));
	/* g_print("subs for %s: ", arg); */
	/* printlist(subs); */
	
	if (subs == NULL)
	{
	    rep_option = is_rep_option(arg) ? arg : NULL;
	    newargslist = slist_push(newargslist, arg);
	}
	else
	{
	    if (rep_option != NULL)
	    {
		/* We need to erase our previous repeated option
		   because it has already been added to the argument
		   list earlier in the processing, but we'll
		   re-add a bit below
		*/
		newargslist = slist_pop(newargslist, NULL);
	    }
	    
	    for (el = subs; el != NULL; el = g_slist_next(el))
	    {
		argslist = slist_push(argslist, el->data);
		
		if (rep_option != NULL)
		{
		    argslist = slist_push(argslist, rep_option);
		}
	    }
	}
	
	g_slist_free(subs);
	/* g_print("argslist: "); */
	/* printlist(argslist); */
	/* g_print("newargslist: "); */
	/* printlist(newargslist); */
    }
    
    return g_slist_reverse(newargslist);
}


void exec_it(void)
{
    int i;
    char **newargs;
    GSList *argslist;
    
    /* print_dump(); */
    
    argslist = replace_args(argv_to_gslist(global_argv + optind+1));
    argslist = g_slist_prepend(argslist, global_argv[optind]);
    
    newargs = (char **)gslist_to_array(argslist)->data;
    
    if (debug)
    {
	g_print("%s", g_shell_quote(newargs[0]));
	
	for (i = 1; newargs[i] != NULL; i++)
	{
	    g_print(" %s", g_shell_quote(newargs[i]));
	}
	g_print("\n");
	exit(EXIT_SUCCESS);
    }
    
    xexecvp(global_argv[optind], newargs);
}


GSList *argv_to_gslist(char **argv)
{
    int i;
    GSList *list = NULL;
    
    for (i = 0; argv[i] != NULL; i++)
    {
	list = g_slist_prepend(list, argv[i]);
    }
    
    return g_slist_reverse(list);
}



GArray *gslist_to_array(GSList *list)
{
    GSList *el;
    GArray *array;
    
    array = g_array_sized_new(TRUE, FALSE, sizeof(list->data),
			      g_slist_length(list));
    
    for (el = list; el != NULL; el = g_slist_next(el))
    {
	array = g_array_append_val(array, el->data);
    }
    
    return array;
}



/* get the GSList of substituted args for arg */
/* return NULL if there are no subs */
GSList *get_subs(const char *alias)
{
    GSList *aliases_el;
    
    for (aliases_el = aliases; aliases_el != NULL;
	 aliases_el = g_slist_next(aliases_el))
    {
	/* remember, the first element's data of each
	   sub-list is the name of the alias */
	/*
	g_print("alias is %s\n", alias);
	g_print("data is %s\n", (char *)((GSList *)aliases_el->data)->data);
	*/
	if (strcmp(alias, ((GSList *)aliases_el->data)->data) == 0)
	{
	    return g_slist_next(aliases_el->data);
	}
    }
    
    return NULL;
}


gboolean is_rep_option(const char *arg)
{
    GSList *slist_el;

    for (slist_el = rep_options; slist_el != NULL;
	 slist_el = g_slist_next(slist_el))
    {
	if (strcmp(arg, slist_el->data) == 0)
	{
	    return TRUE;
	}
    }
    
    return FALSE;
}



void print_dump(void)
{
    g_print("context: %s\naliases:\n", desired_context);
    g_slist_foreach(aliases, gfunc_innerprinter, NULL);
    g_print("repeated options:\n");
    printlist(rep_options);
}


void printlist(GSList *list)
{
    g_slist_foreach(list, gfunc_printer, NULL);
    g_print("\n");
}


/* designed for printing each item of a list */
void gfunc_printer(gpointer data, gpointer user_data)
{
    g_print("%s ", (gchar *)data);
}


/* designed for printing a nested list */
void gfunc_innerprinter(gpointer data, gpointer user_data)
{
    printlist(data);
}


GSList *get_default_rep_options(const char *name)
{
    GSList *list;
    list = NULL;
    
    if (strcmp(name, "gnupg") == 0)
    {
	list = g_slist_append(list, g_strdup("-r"));
	list = g_slist_append(list, g_strdup("--recipient"));
    }
    else
    {
	g_warning("unknown set of defaults: %s", name);
    }
    
    return list;
}


void xexecvp(const char *file, char *const argv[])
{
    execvp(file, argv);
    g_critical("error executing %s: %s", file, g_strerror(errno));
    g_assert_not_reached();
}


void dummy_log(const gchar *log_domain, GLogLevelFlags log_level,
	       const gchar *message, gpointer user_data)
{
}


GSList *slist_pop(GSList *list, void **popped)
{
    if (popped != NULL)
    {
	*popped = list->data;
    }
    
    return g_slist_remove(list, list->data);
}


void cfg_parse_error(const char *s)
{
    g_warning("config parse error: %s line %d: %s",
	      cfg_fullpath->str, cfg_lineno, s);
    return;
}


void cfg_warning(const char *s)
{
    g_warning("config warning: %s line %d: %s",
	      cfg_fullpath->str, cfg_lineno, s);
}


void handle_line(GString *s)
{
    gint    argcp;
    gchar **argvp;
    GError *err = NULL;
    
    if (g_shell_parse_argv(s->str, &argcp, &argvp, &err))
    {
	handle_config_argv(argcp, argvp);
	g_strfreev(argvp);
    }
    else
    {
	if (!((err->domain == G_SHELL_ERROR)
	      && (err->code == G_SHELL_ERROR_EMPTY_STRING)))
	{
	    g_warning("error parsing %s line %d: %s",
		      cfg_fullpath->str, cfg_lineno, err->message);
	}
    }
    
    g_clear_error(&err);
}


void handle_config_argv(int argc, char **argv)
{
    if (argc == 0)
    {
	return;
    }
    
    if (strcmp(argv[0], "context") == 0)
    {
	if (in_desired_context)
	{
	    /* we just ended our desired context, so we're ready to rock */
	    exec_it();
	    g_assert_not_reached();
	}
	
	if (argc == 2)
	{
	    in_context = TRUE;
	    in_desired_context = (strcmp(argv[1], desired_context) == 0);
	}
	else
	{
	    cfg_parse_error("context directive should have exactly one argument, its name");
	}
	
	return;
    }
    
    if (! in_desired_context)
    {
	return;
    }
    
    if (strcmp(argv[0], "alias") == 0)
    {
	if (argc > 2)
	{
	    aliases = g_slist_prepend(aliases,
				      argv_to_gslist(g_strdupv(argv+1)));
	}
	else
	{
	    cfg_parse_error("alias directive needs at least two arguments, its name and one value");
	}
    }
    else if (strcmp(argv[0], "defaults") == 0)
    {
	if (argc == 2)
	{
	    rep_options = g_slist_concat(rep_options,
					 get_default_rep_options(argv[1]));
	}
	else
	{
	    cfg_parse_error("defaults directive needs exactly one argument, its name");
	}
    }
    else if (strcmp(argv[0], "repeat") == 0)
    {
	if (argc >= 2)
	{
	    rep_options = g_slist_concat(rep_options,
					 argv_to_gslist(g_strdupv(argv+1)));
	}
	else
	{
	    cfg_warning("repeat directive has no arguments");
	}
    }
    else
    {
	cfg_parse_error("unknown directive");
    }
}
